/* 
 * cubeSQLServer.java
 * Author: Rick Hofstede, University of Twente <r.j.hofstede@utwente.nl>
 */

import java.io.*;
import java.net.*;
import java.security.*;
import java.util.*;
import org.apache.commons.codec.binary.Base64;
import com.google.gson.*;

public class cubeSQLServer {
	
	/*
	Error codes:
		0: no error
		1: IP address of the host could not be determined
		2: I/O error occured when creating the socket
		3: I/O error occured while reading data
	*/
	private int errorCode;
	private String errorMessage;
	private Socket socket = null;
	private PrintWriter out = null;
	private BufferedReader in = null;
	private Gson gson = new Gson();
	
   /**
	 * Connects to the cubeSQL server.
	 * @param host Host name of the server.
	 * @param port Port number at which the server is listening.
	 * @param username Username.
	 * @param password Password.
	 * @return true on successful connection and authentication, otherwise false.
	 */	
	public boolean connect (String host, int port, String username, String password) {
		resetError();
		
		try {
			socket = new Socket(host, port);
			out = new PrintWriter(socket.getOutputStream(), true);
			in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
		} catch (UnknownHostException ex) {
			errorCode = 1;
			errorMessage = "IP address of the host (" + host + ":" + port + ") could not be determined.";
			System.out.println(errorMessage);
			return false;
		} catch (IOException ex) {
			errorCode = 2;
			errorMessage = "Could not connect to " + host + ":" + port + ". Please check whether cubeSQL Server is running.";
			System.out.println(errorMessage);
			return false;
		}
		
		System.out.println("Connected to cubeSQL on " + host + ":" + port);
		
		String randomNumber = "";
		Random randomGen = new Random();
		while (randomNumber.length() <= 10) {
			randomNumber += randomGen.nextInt(10);
		}
		
		long randpool = Long.parseLong(randomNumber);

		// Compute sha1(randomNumber;username) in HEX mode
		String sha1Username = sha1(randpool + username);

		// Compute sha1(randomNumber;base64(sha1(sha1(password)))) in raw mode
		byte[] sha1PasswordTemp = sha1Raw(sha1Raw(password.getBytes()));
		String sha1Password = sha1(randpool + new String(Base64.encodeBase64(sha1PasswordTemp)));
		
		System.out.println("Trying to authenticate user '" + username + "'...");
		
		// Create and send JSON request
		JsonObject jsonObject = new JsonObject();
		jsonObject.addProperty("command", "CONNECT");
		jsonObject.addProperty("username", sha1Username);
		jsonObject.addProperty("password", sha1Password);
		jsonObject.addProperty("randpool", randpool);
		
		try {
			// Set a timeout on the socket since cubeSQL server is not returning 
			// anything after successful authentication
			socket.setSoTimeout(500);
			sendRequest(gson.toJson(jsonObject));
			socket.setSoTimeout(0); // Infinite timeout
		} catch (SocketException ex) {}
		
		if (isError()) {
			disconnect();
		} else {
			System.out.println("Successfully authenticated user '" + username + "'");
		} 
		
		return !isError();
	}
	
   /**
	 * Disconnects from the cubeSQL server.
	 * @return true on successful disconnection, otherwise false.
	 */	
	public boolean disconnect () {
		// Create and send JSON request
		JsonObject jsonObject = new JsonObject();
		jsonObject.addProperty("command", "DISCONNECT");
		
		try {
			// Set a timeout on the socket since cubeSQL server is not returning 
			// anything after successful authentication
			socket.setSoTimeout(500);
			sendRequest(gson.toJson(jsonObject));
			socket.setSoTimeout(0); // Infinite timeout
		} catch (SocketException ex) {
		}
		
		if (!isError()) {
			System.out.println("Successfully disconnected from cubeSQL server");
		}
		
		try {
			out.close();
			in.close();
			socket.close();
		} catch (IOException ex) {
		}
		
		return !isError();
	}
	
   /**
	 * Executes a SQL query.
	 * @param sql SQL Query.
	 * @return JsonElement, or null on error.
	 */	
	public JsonElement execute (String sql) throws cubeSQLServerException {
		if (!isConnected()) {
			throw new cubeSQLServerException("Not connected to cubeSQL Server.");
		}
		
		resetError();
		
		JsonObject jsonObject = new JsonObject();
		jsonObject.addProperty("command", "EXECUTE");
		jsonObject.addProperty("sql", sql);
		
		return sendRequest(gson.toJson(jsonObject));
	}
	
   /**
	 * Executes a SQL SELECT query.
	 * @param sql SQL SELECT Query.
	 * @return JsonElement, or null on error.
	 */	
	public JsonElement select (String sql) throws cubeSQLServerException {
		if (!isConnected()) {
			throw new cubeSQLServerException("Not connected to cubeSQL Server.");
		}
		
		resetError();
		
		JsonObject jsonObject = new JsonObject();
		jsonObject.addProperty("command", "SELECT");
		jsonObject.addProperty("sql", sql);
		
		return sendRequest(gson.toJson(jsonObject));
	}	
	
	public boolean isError () {
		return (errorCode != 0);
	}
	
	private void resetError () {
		errorCode = 0;
		errorMessage = "";
	}
	
	public boolean isConnected () {
		return (socket != null && socket.isConnected());
	}
	
   /**
	 * Sends a JSON request to the cubeSQL server.
	 * @param jsonRequest String containing encoded JSON request.
	 * @return JsonElement, or null on error.
	 */
	private JsonElement sendRequest (String jsonRequest) {
		out.println(jsonRequest); // Send request
		
		char[] buffer = new char[8192];
		int lineLength = 0;
		String line = "";
		JsonElement element = null;
		JsonParser parser = new JsonParser();
		
		while (true) {
			try {
				lineLength = in.read(buffer, 0, 8192);
				line = new String(buffer, 0, lineLength);
				
				try {
					if (line != null) { // Can be null upon server-side connection termination
						// Try to parse to find out whether complete JSON object has been received
						element = parser.parse(line);
						if (element.isJsonObject()) {
							// Messages with errorCode/errorMessage are always JSONObjects
							JsonObject jsonObject = (JsonObject) element;
							errorCode = jsonObject.get("errorCode").getAsInt();
							errorMessage = (jsonObject.get("errorMsg") == null) ? "" : jsonObject.get("errorMsg").getAsString();
						}

						if (isError()) {
							System.out.println("An error has occurred (" + errorCode + "): " + errorMessage);
						}
					}
					break;
				} catch (JsonParseException ex) {
					System.out.println("JsonParseException: " + ex);
				}
			} catch (SocketTimeoutException ex) {
				break;
			} catch (IOException ex) {
				errorCode = 3;
				errorMessage = "I/O error occured while reading data";
				break;
			}
		}
		
		return element;
	}
	
   /**
	 * Generates a SHA-1 hash from the specified String.
	 * @param data String to be hashed.
	 * @return SHA-1 hash as a String, or original String on error.
	 */
	private static String sha1 (String data) {
		byte[] byteData = sha1Raw(data.getBytes());

		StringBuffer hexString = new StringBuffer();
	   	for (int i = 0;i < byteData.length; i++) {
	   		String hex = Integer.toHexString(0xFF & byteData[i]);
	  	     	if (hex.length() == 1) hexString.append("0");
	  	     	hexString.append(hex);
	   	}
		
		return hexString.toString();
	}
	
   /**
	 * Generates a SHA-1 hash from the specified String.
	 * @param data String to be hashed.
	 * @return SHA-1 hash as raw byte code.
	 */	
	private static byte[] sha1Raw (byte[] data) {
		byte[] result = null;
		
		MessageDigest digest;
		try {
			digest = MessageDigest.getInstance("SHA-1");
			digest.update(data);
			result = digest.digest();
		} catch (java.security.NoSuchAlgorithmException ex) {
		}
		
		return result;
	}
	
}
